/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Marcell Gal cell at x-dsl.hu]                *
***********************************************************
* This file is included, not linked
* Originally used for EDIS
* Now also used for driving dummier ign modules
* also our ign259 module onboard genboard v3.x
\*********************************************************/

/*
definitely (from ECM point of view) EDIS is the simplest way of ignition:
it just needs a small-current SAW signal
that goes active at the falling edge of the PIP and goes inactive
(1536 - (25.6*engine.ignadv >> 2)) usec afterwards. Note that engine.ignadv
is currently declared as 1/4 degree resolution, therefore the >>2.

You can add SAW += 2048 usec
below 1200 RPM if multispark is enabled in config.
*/

#ifndef __EDIS_C__
#define __EDIS_C__
#include "edis.h"
#include "ve.h"
#include "arithm.c"
#include "lbus.c"

#ifndef NOIGN
/* polarity depends on config 
Dispatcher uses these just like trigger.c).
Naming comes from what the IGBT does in case of dummyignition.

TODO: #ifdef GENBOARDv3 
control via the p259 (allocate a pin) or ign259 depending on config
*/
#ifdef GENBOARDv3
// currently only a single IO pin
/*!
 * \ingroup g_edis_timing
 * \brief Activate IGN_OUT bit - Ver 3 Genboard.
 * MUST be called under cli()
 * \return none
 */

#define ignport_map(x) (hwmap.table[2*HWMAP_SIZE+x])
 
static inline void ign_activate(void){
	uint8_t ch, sign;
	uint8_t enignct = engine.igncount;
	ch = ignport_map(enignct);
	sign = config.ign_out & _BV(IGN_INVERTOUT);
	if(sign) {
		if(enignct != 0xFF) {
			I259(DLOW, ch);
#ifdef IGN_DUALOUT
		// would it make sense to do another lookup ch = ignport_map( f(enignct) );
			I259(DLOW, ch-1);
#endif
		}
#ifdef V3_IGN_TEST
		PORTD &= ~_BV(CS_CAN);
#endif
	} else {
		if(enignct != 0xFF) {
			I259(DHIGH, ch);
#ifdef IGN_DUALOUT
			I259(DHIGH, ch-1);
#endif
		}
#ifdef V3_IGN_TEST
		PORTD |= _BV(CS_CAN);
#endif
	}
}

/*!
 * \brief De-Activate IGN_OUT bit - Ver 3 Genboard.
 * MUST be called under cli()
 * \return none
 */
static inline void ign_deactivate(void){
	uint8_t ch, sign;
	uint8_t enignct = engine.igncount;
/* this condition could only set a variable to DHIGH or DLOW
 * However by using constants, maybe gcc makes some optimizations,
 * using slightly more flash
 */
	ch = ignport_map(enignct);
	sign = config.ign_out & _BV(IGN_INVERTOUT);
	if(sign) {
		if( enignct != 0xFF) {
			I259(DHIGH, ch);
#ifdef IGN_DUALOUT
			I259(DHIGH, ch-1);
#endif
		}
#ifdef V3_IGN_TEST
		PORTD |= _BV(CS_CAN);
#endif
	} else {
		if(enignct != 0xFF) {
			I259(DLOW, ch);
#ifdef IGN_DUALOUT
			I259(DLOW, ch-1);
#endif
		}
#ifdef V3_IGN_TEST
		PORTD &= ~_BV(CS_CAN);
#endif
	}
}

// reset ignchannels
static inline void ign_deactivate_all(void) {
	uint8_t i, dhigh_or_dlow;

	dhigh_or_dlow = (config.ign_out & _BV(IGN_INVERTOUT)) ? DHIGH : DLOW ;
	i = 8;
	while( i ){
		i--;
		I259(dhigh_or_dlow, ignport_map(i));	
	}
}

#ifdef V3_IGN_DRIVE_A
/*
 Some functions for testing hardware activated ign output
 One of the eventqueue output compare units is used
*/

static inline void test_ign_activate_at(uint16_t t) {
	if(config.ign_out & _BV(IGN_INVERTOUT)) {
		TCCR1A |= _BV(COM1A1);
		TCCR1A &= ~_BV(COM1A0);
	} else {
		TCCR1A |= _BV(COM1A1) | _BV(COM1A0);
	}
	OCR1A = t;
}
	
static inline void test_ign_deactivate_at(uint16_t t) {
	if(config.ign_out & _BV(IGN_INVERTOUT)) {
		TCCR1A |= _BV(COM1A1) | _BV(COM1A0);
	} else {
		TCCR1A |= _BV(COM1A1);
		TCCR1A &= ~_BV(COM1A0);
	}
	OCR1A = t;
}

// force the output compare to take effect
static inline void test_ign_action_now(void) {
	TCCR1C = _BV(FOC1A);
}

static inline void test_ign_turnoff(void) {
	test_ign_deactivate_at(0);
	test_ign_action_now();
}

static inline void test_ign_turnon(void) {
	test_ign_activate_at(0);
	test_ign_action_now();
}

#endif

#else // not GENBOARDv3

/*!
 * \ingroup g_edis_timing
 * \brief Activate IGN_OUT bit.
 *
 * \return none
 */
static inline void ign_activate(void){
	if(config.ign_out & _BV(IGN_INVERTOUT))
		IGN_PORT &= ~_BV(IGN_OUT);
	else
		IGN_PORT |= _BV(IGN_OUT);
}

/*!
 * \brief De-Activate IGN_OUT bit.
 *
 * \return none
 */
static inline void ign_deactivate(void){
	if(config.ign_out & _BV(IGN_INVERTOUT))
		IGN_PORT |= _BV(IGN_OUT);
	else
		IGN_PORT &= ~_BV(IGN_OUT);
}

static inline void ign_deactivate_all(void) {
	// empty for v2
}
#endif	// not GENBOARDv3

// EDIS could be dropped from my_make..

/* this pulls the SAW signal high. 
 * BEWARE: if you use an inverting output, like misc NPN with
 * a pullup resistor, this must pull the AVR-pin low
 */

/*!
 * \ingroup g_edis_timing
 * \brief Initialize the IGN_OUT bit.
 *
 * \return none
 */
static inline void ign_init(void)
{
	
#ifdef GENBOARDv3
	/*	The ign259 has a HW reset circuit.
		Should we explicitely go through the channels,
		so we save the coils if power outage was partial,
		only resetting AVR?
		HC259 can live downto 2V.
	*/
	
	ign_deactivate_all();

#ifdef V3_IGN_DRIVE_A
	test_ign_turnoff();
#endif

	return;
#else // GENBOARD v2.2
	IGN_DDR |= _BV(IGN_OUT); 	// set port to output
#ifdef __MEGASQUIRT_C__
	ign_deactivate();
#else
// arghhh.:
// ioinit() when used from bootloader does not know about config
// this could be bridged by perl, since make_conf.pl knows about ign_out bit0
// either reedit my_make or generate a file which we include
#ifdef IGN_OUT_INVERTED
	IGN_PORT |= _BV(IGN_OUT);
#else
	IGN_PORT &= ~_BV(IGN_OUT);
#endif // IGN_OUT_INVERTED

#endif // __MEGASQUIRT_C__
#endif // GENBOARD v2.2
}

#include "eventqueue.h"

/* 
Called from userspace:
calculates SAW length, which will be put to engine.dwell.

What if needed saw is very short (cause user config is stupid)
Who configures 60 degrees igniton advance with EDIS? I hope noone!
*/
/*!
 * \ingroup g_edis_timing
 * \brief calculates SAW length, which will be put to engine.dwell.
 *
 * \return saw_length  length of the EDIS SAW output
 */
static inline uint16_t edis_sawlen(void)
{
	uint16_t t, saw_length;
	t = mult16_8(410, engine.ignadv)>>8;	// 410 comes from: 25.6*16
	// would it be faster with *205 (8bit * 8bit) and >>7 ? probably not.
#define MAX_SAW_LEN (1536/4)
#define MIN_SAW_LEN (100/4)
	if((MAX_SAW_LEN - MIN_SAW_LEN) < t)
		saw_length = MIN_SAW_LEN;
	else
		saw_length = MAX_SAW_LEN - t;

	// += 2048 usec for when RPM<1200 and multispark enabled in config
	// around 1100 RPM we are over-cautios
	if((engine.rpm < 11) && (config.ign_out & _BV(IGN_MULTISPARK)))
		return saw_length + 512;

	return saw_length;
}
#endif // NOIGN
#endif // __EDIS_C__

