/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Michael Kristensen, mik@caffrey.dk]          *
***********************************************************
*  -- History --                                          *
*    release 0.1 - 06-03-2003                             *
*            http://caffrey.dk/megasquirt                 *
*    initial Motorola version by Bowling & Grippo (2002)  *
*            http://www.bgsoflex.com/megasquirt.html      *
\*********************************************************/


/*
TODO:

There is a timeout during acceleration, this should be fixed.
Is the same timeout appearing in the B&G asm code?

*/

   /*!
   * \addtogroup g_fuel_calc Fuel Calculation Routines
   */
/*@{*/

#include "global.h"
#include "fuelcalc.h"
#include "fuelcontrol.h"
#include "helpers.h"
#include "benchmark.h"
#include "adc.h"
#include "ve.h"
#include "tables.h"
#include "eventqueue.h"
#include "injconf.h"
#include "ego.h"
#include "lcd.h"

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <avr/pgmspace.h>
#include <string.h>
#include <avr/delay.h>
#include "arithm.c"

volatile extern struct engine_t engine;

uint8_t tpsaclk; // TPS enrichment timer clock in 0.1 second resolution

uint8_t tpsaclkcmp; // Comparison value for TPS acceleration time - from lookup table

struct corr_t corr;
struct search_table_t clt_idx;

#include "injector.c"

/*
	progressive_delay() allows configuration of higher 
	than 25.5 msec in 8 bit config value. Used for config.cwl, cwh, injpwmt, primep
	Value 200 mean 20msec (megatune compatibility 100 usec step)
	but 201 means 20.5 msec (changing to 500 usec step)
	This allows 20 msec + 55 * 0.5 msec = 47.5 msec max pulsewidth
*/
#define PROGRESSIVE_FUNCTION_BREAK (200)
static inline uint16_t progressive_delay(uint8_t t){
	uint16_t ret;
	if(t <= PROGRESSIVE_FUNCTION_BREAK){
		return 25 * t;	// 100 usec step
	}
	
	ret = PROGRESSIVE_FUNCTION_BREAK * 25;
	t -= PROGRESSIVE_FUNCTION_BREAK;
	
	ret += t * 125; // 500 usec step
	return ret;	
}

/*
 * ONLY call from userspace
 * many squirt_t-s to fill (called from several places)
 */
void fill_all_squirters(uint16_t pw)
{
	uint8_t i, mask;
	uint16_t t;
	struct squirt_t *inj_port_p;
	// config.injpwmt is 100usec based, multiply by 25 to make it 4usec based
	inj_port_p = &inj_port[0];
	t = progressive_delay(config.injpwmt);
#ifdef INJECTOR_STAGING
	i = INJECTORS-SECONDARY_INJ_CHANNELS_NUM;
#else
	i = INJECTORS;
#endif //INJECTOR_STAGING
	do{
		i--;
		cli();
		inj_port_p->pwm_delay = t;
		inj_port_p->pwcalc = pw;
		sei();
		inj_port_p++;
		// TODO: check if gcc adds sizeof...
	} while(i);

#ifdef INJECTOR_STAGING
	// Calculating pulsewidth of secondary injectors
	uint16_t pw2;
	if(engine.tps >= config.inj_stage2_start_tps &&
		engine_kpa8() >= config.inj_stage2_start_map)
		pw2 = mult16_16(pw, (uint16_t)(config.inj_stage2_rate) + 256)>>8; // effectively (256 + 256) / 256
	else
		pw2 = 0;
	i=SECONDARY_INJ_CHANNELS_NUM;
	while(i)
	{
		i--;
		cli();
		inj_port_p->pwm_delay = t;
		inj_port_p->pwcalc = pw2;
		sei();
		inj_port_p++;
		// TODO: check if gcc adds sizeof...
	}
#endif	

	// calculate injector bitmask used when firing all injectors simultaneously
	i = config.alternate & 0x0f;
	mask = injport_map(i);
	while(i) {
		i--;
		mask |= injport_map(i);
	}
	engine.injector_bitmask = mask;
}

void init_fuelcalc(void) {

  corr.ego = EGOCORR_BASE;
  corr.air = 100;
  corr.warm = 0;
  corr.baro = 100;
  corr.gammae_1k5 = GAMMAE_BASE;
  corr.ve_hp = 25600;
  corr.lambda = LAMBDACORR_STOICH;
  corr.tpsaccel = 0;
  corr.tpsfuelcut = 100;
  corr.tps_overrun = 0;
  corr.als = 0;

	fill_all_squirters(0);

  // you will have a small few msec window at about 6 hours after powerup to
  // generate fake rpm for the first trigger:
  etime.ptrig_capturedtime_prev |= 0xFF000000; // gcc: 1 byte, please
  //engine.rpm_trigger_reenable = TRIGGER_REENABLE_MIN;
}

// search the warmup_clt_range table
// TODO: only run this when engine.coolant changes
void search_clt_index(void) {
  // note, coolant is actual temperature +40 degrees F
  search_table(config.warmup_clt_range, sizeof(config.warmup_clt_range), engine.coolant, &clt_idx);
}

void leave_cranking(void)
{
	uint8_t my_status;
	cli();
	my_status = engine.status;
	if (my_status & (uint8_t) _BV(crank)) {
		my_status &= ~_BV(crank);
		my_status |= _BV(startw) | _BV(warmup);
		engine.status = my_status;
		etime.asecount = config.awc;
	}
	sei();
	
	if(config.lcd_c0 & _BV(3)){
		// start another initialization, maybe low voltage at 
		// cranking resetted the LCD
		lcd_initfrom(1);
	}
}

/***************************************************************************/
/*** warm-up and after-start enrichment                                  ***/
/*** corr.warm = 0 means *1.0 no enrichment (old 100)                    ***/
/*** corr.warm=155 means *2.55 (old max: 255)                            ***/
/*** corr.warm=255 means *3.55 (not possible before)                     ***/
/***************************************************************************/
void warmup_enrich(void) {
	uint8_t afterstart;
	uint8_t w8;	// warmup enrichment 8 bit value biased at 0
	uint16_t w;	// warmup enrichment + afterstart final calc, biased at 0

	// afterstart enrichment
	afterstart = 0;
	if (engine.status & _BV(startw)) {
		if (etime.asecount) {
			uint8_t ase_cold, ase_warm;

			ase_cold = linear_interp(0, config.awc, 0, config.awev, etime.asecount);
			ase_warm = (ase_cold * config.awev_temp_scaling) >> 8;
			afterstart = linear_interp(COLD_TEMP, WARM_TEMP, 
							ase_cold, ase_warm, engine.coolant);
		} else {
			cli();
			engine.status &= ~_BV(startw);
			sei();
		}
	}

	// no need to run through everything if the engine is warm.
	// this results in the value of wwu[last_element]is ignored.
	if (engine.coolant >= config.warmup_clt_range[sizeof(config.warmup_clt_range)-1]) {
		// outside warm-up range, clear warm-up enrichment code
		cli();
		engine.status &= ~_BV(warmup);
#ifndef GENBOARDv3
		ledstate &= ~_BV(LED_WARMUP); // wled off
#endif
		sei();

		corr.warm = afterstart;
		return;

	} else {
		cli();
#ifndef GENBOARDv3
		ledstate |= _BV(LED_WARMUP); // wled on
#endif
		engine.status |= _BV(warmup);
		sei();

		// enrichment due to coolant
		w8 = linear_interp(config.warmup_clt_range[clt_idx.lindex],
				  config.warmup_clt_range[clt_idx.uindex],
				  config.warmup_clt[clt_idx.lindex],
				  config.warmup_clt[clt_idx.uindex],
				  engine.coolant);

		// the semantics is traditionally stupid in megatune:
		// w8 is biased at 100 at this point, so we make it biased 0:
		if( w8 > 100 ){
			w8 -= 100;
#warning "warmup_enrich semantics changed to simpler config !"
		// only possible to downscale
		// the (warmup_clt[] - 100) part, with increasing rpms
		w8 = linear_interp(ve.rpm_range[0], ve.rpm_range[VE_SIZE_RPM-1], 
				w8, (w8*config.warmup_rpm_scale)>>8, engine.rpm);

		/* OBSOLETE, multiple-bins enrichment due to rpm:
		rpm_enrich = linear_interp(ve.rpm_range[rpm_idx.lindex],
						ve.rpm_range[rpm_idx.uindex],
						config.warmup_rpm[rpm_idx.lindex],
						config.warmup_rpm[rpm_idx.uindex], 
						engine.rpm );
		*/
		} else {
			// no enrichment (value=100) or stupid config (low-CLT leaning not applied)
			w8 = 0;
		}
	}
		
	// add the afterstart enrichment and check for overflow
	w = w8 + afterstart;

	// check overflow:
	corr.warm = (w >> 8) ? 0xFF : (uint8_t)w;
}

/*
 * overrun fuelcut
 * return 1 if fuelcut is required
 */
uint8_t tps_overrun(void) {

	// Buck-free fuel saver!
	// When throttle is fully released the injectors will be turned off as long
	// as rpm >= fuelcut-rpm.
	// The injectors are turned on again when rpm <= fuelresume-rpm.
	// suggested values:
	//   fuelresume = 0x10 (1600 rpm)
	//   fuelcut = 0x11 (1700 rpm)

	// remember to specify config.iac_tps_thres else this won't work!
	if (engine.rpm <= config.overrun_fuelresume)
		corr.tps_overrun &= ~_BV(OVERRUN_ENABLE);
	
	else if (engine.rpm >= config.overrun_fuelcut)
		corr.tps_overrun |= _BV(OVERRUN_ENABLE);


	if ( engine.tps > config.iac_tps_thres ){
		if( debug.disable_flags & _BV( DIS_DONT_MOVE ) ){
			// shut down the engine if someone steps on the TPS (antitheft)
			return 1;
		}

		// disable overrun when tps is pressed
		corr.tps_overrun &= ~_BV(OVERRUN_ACTIVE);
		return 0;
	} else {
		// TPS not pressed
		if(etime.asecount != 0) {
			// don't do any fuelcut during afterstart
			corr.tps_overrun &= ~_BV(OVERRUN_ACTIVE);
		        return 0;
		} else if (corr.tps_overrun & _BV(OVERRUN_ENABLE)) {
			// throttle released and rpm still way above idle
			corr.tps_overrun |= _BV(OVERRUN_ACTIVE);
			return 1;
		} else {
			corr.tps_overrun &= ~_BV(OVERRUN_ACTIVE);
			return 0;
		}
	}
}


/***************************************************************************/
/*** Throttle position acceleration enrichment                           ***/
/***************************************************************************/
void tps_acc_enrich(void) {
  uint8_t tok_diff; // tok = tps or kpa
  uint8_t my_tok, my_last_tok;
  
  cli();
  if (config.tpsdot_kpadot_conf & _BV(tps_or_kpa)) my_tok= engine_kpa8();
  else my_tok= engine.tps;
  my_last_tok = engine.last_tok;
  sei();

  if (my_tok < my_last_tok) {
    tok_diff = my_last_tok - my_tok;
  
    /* deceleration */
    if (tok_diff >= config.tps_thresh) {  // above threshold?
      if (engine.status & _BV(tpsaen)) {
	corr.tpsfuelcut = 100;
	corr.tpsaccel = 0;
	cli();
	engine.status &= ~( _BV(tpsaen) | _BV(tpsden) );
#ifndef GENBOARDv3
	ledstate &= ~_BV(LED_ACCEL); //aled off
#endif
	sei();
      } else { // fuel cut
	if (engine.rpm >= config.decel_fuelcut_thres) {
	  corr.tpsfuelcut = config.tpsdq;
	  cli();
	  engine.status &= ~_BV(tpsaen);
	  engine.status |= _BV(tpsden);
#ifndef GENBOARDv3
	  ledstate &= ~_BV(LED_ACCEL); //aled off
#endif
	  sei();
	}
      }
    } else {
      if (engine.status & _BV(tpsden)) {
	cli();
	engine.status &= ~( _BV(tpsaen) | _BV(tpsden) );
	sei();
	corr.tpsfuelcut = 100;
	corr.tpsaccel = 0;
      }
    }


  } else {
    tok_diff = my_tok - my_last_tok;

    /* acceleration */
    if (tok_diff >= config.tps_thresh) {

      /*
	The acceleration calculation needs to be reworked, it is unprecise.
	It times out, and restarts computation of enrichment.
      */

      if (engine.status & _BV(tpsaen)) {
	uint8_t acc_temp_offset, acc_temp_mult, acc_enrich;
	uint16_t mtmp;
	struct search_table_t st;

	/* multiplier amount based on cold temperature */
	acc_temp_mult = linear_interp( COLD_TEMP, WARM_TEMP, config.acmult, 100, engine.coolant );
	
	/* lookup table amount based on tpsdot */
	search_table(config.tpsdotrate, sizeof(config.tpsdotrate), tok_diff, &st);

	acc_enrich = linear_interp(config.tpsdotrate[st.lindex],
				   config.tpsdotrate[st.uindex],
				   config.tpsaq[st.lindex],
				   config.tpsaq[st.uindex],
				   tok_diff);

	//	acc_enrich = linear_interp(st.lbound, st.ubound, config.tpsaq[st.index-1], config.tpsaq[st.index], tps_diff);

	/* calculate acc-enrichment, catch overflow */
	mtmp = mult_div100(acc_temp_mult, acc_enrich);
	if ((mtmp >> 8) != 0) // will the result fit in a byte?
	  mtmp = 200;         // ... no!

	/* extra fuel due to temperature */
	acc_temp_offset = linear_interp( COLD_TEMP, WARM_TEMP, config.tpsacold, 0, engine.coolant );

	mtmp += acc_temp_offset;
	if ((mtmp >> 8) != 0) // will the result fit in a byte?
	  mtmp = 0xFE;        // ... no!
	
	if (mtmp > corr.tpsaccel) 
	  corr.tpsaccel = (uint8_t)mtmp; // tpsaccel is the result of this function

	if ((engine.status & _BV(tpsden)) || (tpsaclk >= tpsaclkcmp))  {
	  cli();
	  engine.status &= ~( _BV(tpsaen) | _BV(tpsden) );
#ifndef GENBOARDv3
	  ledstate &= ~_BV(LED_ACCEL); //aled off
#endif
	  sei();
	  corr.tpsfuelcut = 100;
	  corr.tpsaccel = 0;
	}
      } else { //start of acceleration, initialize some variables
	corr.tpsaccel = config.tpsaq[0];
	tpsaclk = 0;
	tpsaclkcmp = config.tpsasync;
	cli();
	engine.status |= _BV(tpsaen);
	engine.status &= ~_BV(tpsden);
#ifndef GENBOARDv3
	ledstate |= _BV(LED_ACCEL); //aled on
#endif
	sei();
	//	ledstate &= ~_BV(LED_WARMUP); //wled off
      }
    } else { //is acceleration done?
      if ((engine.status & _BV(tpsden)) || (tpsaclk >= tpsaclkcmp)) {
	cli();
	engine.status &= ~( _BV(tpsaen) | _BV(tpsden) );
#ifndef GENBOARDv3
	ledstate &= ~_BV(LED_ACCEL); //aled off
#endif
	sei();
	corr.tpsfuelcut = 100;
	corr.tpsaccel = 0;
      }
    }
  }
}

/* 
* if pw is bigger than the pwm_delay, 
* we are called to calculate the
* pwm percentage. The right way to calculate pwm percentage is
* from VBatt (similar to dwell), dispite some guys calculate
* it from rpm (cranking => low Batt), which is ugly.
* Simple interpolation according to VBatt.
* note that output range is [ v14, v14+v6 ]
* for battery voltage [14V, 6V]
* result is trimmed to 8 bits.
*/
static inline uint8_t calc_pwm_dc(uint8_t v14, uint8_t v6)
{
	uint8_t vbatt;
	uint16_t dc;

	// for anything above 14V we use the v14 value
	// (not decreased for higher than 14V voltages)
	// this is very convenient, the user normally need to set only 1 value
	dc = v14 ;
	vbatt = engine.batt;
	// TODO: use the defines, 164 is 19V!!
	// 164 (appr 14V?) and 61 (6V?) ADC readings seem to be typical high-low values
	if(vbatt<164) {	// lower Battery voltage, correction applies
		dc += ((164-vbatt) * v6 )>>7;	// at 6V: 103/128 * v6 added to v14
	}
	
	if(dc>>8) return 0xFF;
	return dc & 0xFF;
}

/*
* used from several places
*/
void static inline pwm_dc_fill(void)
{
// genboard v3 hasn't individual dutycycle setting
#ifdef GENBOARDv3
	// #ifdef FETDRIVER_INVERTING taken care at initialization
	OCR0 = calc_pwm_dc(config.injpwm, config.injpwm6);
#else
	if( inj_port[0].pwcalc > inj_port[0].pwm_delay || inj_port[1].pwcalc > inj_port[1].pwm_delay ){
		inj_port[0].pwm_dc = inj_port[1].pwm_dc = calc_pwm_dc(config.injpwm, config.injpwm6);
	}
#endif
}

/*
* Calculates cranking injector pulsewidth in 4 usec
* by doing interpolation according to coolant temperature
* between pulsewidths (progressive function of)
* config.cwl and config.cwh
* 
* Considers TPS FLOOD_CLEAR
*/
static inline uint16_t cranking_pw(void)
{
	if (engine.tps < FLOOD_CLEAR) {
		// calculate pulsewidth in 4 usec from [cwl..cwh] 
		uint16_t cwl16 = progressive_delay(config.cwl);
		uint16_t cwh16 = progressive_delay(config.cwh);
		
		return linear_interp16( COLD_TEMP, WARM_TEMP, cwl16, cwh16, engine.coolant );
		// battery correction?
	} else { 
		return 0; // flood clear
	}
}

/*
* Returns 255 if eng_bat <= minv; smaller than 255 if eng_bat is higher.
* Slope is a parameter.
* example:
* slope=0x80
* eng_bat = BATT_14V (118)
* minv = BATT_7V (61)
* result: 226 (*0.88)

* with slope=0xFF same result=198 (*0.77)
*/
static inline uint8_t vbatt_scale(uint8_t eng_bat, uint8_t slope, uint8_t minv )
{
	if (eng_bat <= minv){
		return 255;
	}

	return 255 - ((slope * (eng_bat - minv)) >> 8);
}

/* 
* Battery voltage compensation ( in 16 usec units )
* Remember, low battery voltage only delays injector opening time
* does not affect flow after inj has opened 
* and does not change closing time with constant-voltage
* (zener) flyback. With DR flyback or flyrail with
* voltage proportional to VBATT, the closing time is affected too.
* With DR flyback and lowZ the closing is not effected,
* since the pwm-duty (ideally) compensates for VBATT,
* setting a fixed current.
*
* Note that this is just the simpler part of pulsewidth extension:
* See injrampup_pw_extend()
*/
uint16_t vbatt_pw_extend(void)
{
	uint8_t batt_high, batt_low;
	uint16_t ge;

	ge = config.injopen + config.battfac; // take care of overflow
	if (ge > 0xff)
		batt_low = 0xff;
	else
		batt_low = (uint8_t)ge;

	if (config.injopen <= config.battfac)
		batt_high = 0;
	else
		batt_high = config.injopen - config.battfac;
	
// batt_low, batt_high resolution is 16 usec (max value ~4 ms)
// so config.injopen and config.battfac also (TODO: check megatun.. issues)

	return linear_interp( BATT_7V, BATT_19V, batt_low, batt_high, engine.batt );
}

/* 
* returns pulsewidth extension necessary due to rampup of injector-fuelflow.
* (in 16 usec units).

http://www.vems-group.org/index.php?page=GenBoard%2FManual%2FConfig%2FInjectorOpening

	pw: calculated fuel * req_fuel [4 usec] 
	maxext: [16 usec] is the max value used for extension (at BATT_7V) and high pw.
	For higher battery voltage return value will be downscaled linearly.

	Furthermore, for small pulsewidth, it will be downscaled according to 
	a lookup table. (that holds a very scientific squareroot-like function)
*/
uint8_t injrampup_pw_extend( uint16_t pw, uint8_t maxext){	
	uint8_t maxext_at_vbatt, index, t;

	if( maxext == 0){
		return 0;
	}
	
	// config.injocfuel [16 usec] 
	maxext_at_vbatt = (maxext 
		* vbatt_scale( engine.batt, config.injrampup_battfac, BATT_7V ))>>8;

	/*
		rampup_func_power = 1.5
		if pw is above (2/3 *maxext_at_vbatt)
		which is (maxext_at_vbatt / rampup_func_power) 
		the return value will be exactly maxext_at_vbatt

		Note that we'd use 171/256 not 42/16 if both were 16 usec
	*/
	if( pw >= ((maxext_at_vbatt * 42 )>>4) ){
		return maxext_at_vbatt;
	}

	/* if pw is lower (this is only for very small pulsewidths) 
		lookup a multiplier in a flashlookup table 
		and downscale maxext_at_vbatt accordingly.
		Note that we cannot avoid a division
		to find index (although 1/maxext_at_vbatt could be cached)

		pw canNOT be higher than 1020 (4080 usec) at this point.
	*/
#define INJOPENTABLE_SIZE 64
// or (sizeof(INJOPENTABLE[])) ?
// because of pw and rampup_at_vbatt we only need to multiply with 
#define IOTMULTIPLIER 24
// or (1.5 * INJOPENTABLE_SIZE / 4)
	index = (pw * IOTMULTIPLIER) / maxext_at_vbatt;
	// no need to check index because of above conditions
	t = PRG_RDB( &INJOPENTABLE[ index ] );
	// finally downscale:
	return ( maxext_at_vbatt * t ) >> 8;
}

/* 
* Helper function to calculate injector pulsewidth [4 usec]
* enables forced pulsewidth, 
* considers rev-limiter
* and cranking
*/

uint16_t total_pw(void)
{
	uint16_t res, active_rev_limit;

	// emil asked that config.als_rev_limit can raise rev_limit not just decrease it
	active_rev_limit = is_als_active() ? config.als_rev_limit : config.rev_limit ;
	// rough revlimiter
	if(engine.rpm > active_rev_limit) return 0;

	// mdp menu command can force a pulsewidth in 64 usec resolution
	if(debug.forced_pw) return debug.forced_pw << 4;
	
	if ((engine.status_ext & _BV(crank_enable)) 
		&& (engine.rpm <= config.cranking_thres)) {
		return cranking_pw();
	}  

	// overrun fuelcut
#ifdef ALS
	if (als_deactivated && tps_overrun()) return 0; // we don't want to cut fuel when ALS switch is ON!
#else // No ALS
	if (tps_overrun()) return 0;		
#endif // ALS

	// some multipliers are 100 based, stupid inheritance:
	res = mult16_8((corr.warm + 100), corr.tpsfuelcut);
	// res base is now 10000. Max value is 355*100 (< 65535 fine)
	res = mult16_16(corr.baro * corr.air, res)>>16;	// res is 1526 based
	
	// other multipliers are more reasonable, based on power of 2
#if EGOCORR_BASE == 128
	res = mult16_16(res, (uint16_t)(corr.ego) + 128)>>8; // effectively (256 +- 128) / 256
#else 
#warning "EGOCORR_BASE is broken at least in comm.c and lcd_display"
#endif

#ifdef ALS
    res = mult16_16(res, (uint16_t)(corr.als) + 256)>>8; // effectively (256 + 256) / 256
#endif

	// store high precision gamma-enrichment:
	corr.gammae_1k5 = res;

	/*	up to this point all multipliers were close to 1.0 
		
		Let's apply VE, which is (256 * 100) based:
	*/
	res = div4096(mult16_16(corr.ve_hp, res));
	// res is now based on 9536.7 

	// apply the lambdacorr
	res = mult16_16(res, (uint16_t)(corr.lambda) + LAMBDACORR_OFFSET)>>8; // effectively (300 +- 128) / 256

	// any remaining multiplier constant from above and below (req_fuel!!) is fixed here
	if (!(config.config13 & _BV(CONTROL_STRATEGY))){  // speed-density
		res = mult_div596(engine.kpa250, res);
	} else {
		res = mult16_8(res, 172) >> 8;
	}

	// config.req_fuel is 100 usec based (little clumsy, but we stick to it)
	// we return in 4 usec resolution
	return mult16_8( res, config.req_fuel ) >> 8;
}

/***************************************************************************/
/*** calc total enrichment                                               ***/
/***************************************************************************/
void calc_total_enrichment(void) {
	uint16_t res;
	res = total_pw();
		
	// final pulsewidth calculation, wuhuw!
	if (res) {
		uint8_t batt_tmp;
		// uint16_t oc;

		// note that config.injocfuel semantics is changed
		// also note that this is called _before_ vbatt_pw_extend() applied!
		batt_tmp = injrampup_pw_extend( res, config.injocfuel );
		res += ((uint16_t)batt_tmp<<2);

		batt_tmp = vbatt_pw_extend();
		res += ((uint16_t)batt_tmp<<2) + (corr.tpsaccel*25);

		/*
		oc = ((uint16_t)config.injocfuel<<2);

		// take care of underflow
		if (res > oc)
			res -= oc;
		else
			res = 0;
		*/
	} 

	fill_all_squirters(res);
	pwm_dc_fill();
}

/*
downscale multiplier for sensor calibration 
(can be used for other sensors too - maybe use h[] table for config ?)
TODO: 10 bit ADCreading
TODO: mik, you might want to make this upscale only, or up/down
scale so less resolution is lost with newer,
more precise voltage divider standard 
(VBATT_ ... constants must be matched, and megatune issue?)
*/
static inline uint8_t batt_cal(uint8_t adcreading, uint8_t factor) {

	uint16_t t16;

	t16 = adcreading * factor;
	return (t16 >> 8);
}

// config.kpafac resolution. 2 is default, 4 (kPa) is possible. Only effects kpafac variable
#if KPAFAC_UNIT==4
#define EXPONENT_OF_512_PER_KPAFAC_UNIT 7
#else // KPAFAC_UNIT==4
#if KPAFAC_UNIT==1
#define EXPONENT_OF_512_PER_KPAFAC_UNIT 9
#else // KPAFAC_UNIT==1
	
#if (defined KPAFAC_UNIT) && (KPAFAC_UNIT!=2)
#warning "BAD KPAFAC_UNIT configured. Assuming KPAFAC_UNIT=2"
#endif

// default:
#define EXPONENT_OF_512_PER_KPAFAC_UNIT 8

#endif // KPAFAC_UNIT==1
#endif // KPAFAC_UNIT==4

// kpaoffs range is +- 32kPa
#define KPAOFS_BIAS (128)

/*
 * fill engine.kpa250 from the filtered MAP sensor signal.
 * Normally
 * 0..2046 [*250 Pa]
 *=0..512 kPa (impressive boost !!)
 * but with KPAFAC_UNIT=4
 * maximum is
 * 0..4092
 *=0..1023 kPa
 */
static inline void kpafill(void){
	uint8_t *p;	

	if(config.kpafac != 0) {
		sample_t k = sensors11(MAP);	// max 2046..2047
		if(k == 0 || k >= 2045){	// oops, let's not believe this value
			k = 4 * config.baro;
		} else {
			k = (mult16_8(k, config.kpafac)) >> EXPONENT_OF_512_PER_KPAFAC_UNIT;
			k += config.kpaoffs;
/*
    transfer function V=f(P) of mpx4250 at P=0 intercepts V in -0.04*Vs,
    sometimes config.kpaoffs=-5kPa is desirable
*/

			if( k > KPAOFS_BIAS ){
				k -= KPAOFS_BIAS;
#if EXPONENT_OF_512_PER_KPAFAC_UNIT < 8
	// otherwise railing not needed, ADC values limit MAP < 512 kPa
	// but we don't need railing anyway, this allows control for MAP upto 1023 kPa
	//			if( k > 2047) k = 2047;
#endif
			} else {
				k = 0;
			}
		}
		engine.kpa250 = k;
		return;
	}

	// if you get here, your resolution is motorola megasquirt compatible => sucks!
	// Manifold Air Pressure in kiloPascals
	if (config.config11 & _BV(MAP_SENSOR)){
		p = (void*)KPAFACTOR4250;
	} else {
		p = (void*)KPAFACTOR4115;
	}

    engine.kpa250 = 4 * PRG_RDB(p + sensors8(MAP));
}	// kpafill()

/***************************************************************************/
/*** calc parameters                                                     ***/
/***************************************************************************/
void calc_parameters(void) {
  uint8_t t;

	kpafill();

  // Coolant temperature in degrees F + 40
  engine.coolant = PRG_RDB(&THERMFACTOR[sensors8(CLT)]);

  // Manifold temperature in degrees F + 40
  engine.mat = PRG_RDB(&MATFACTOR[sensors8(MAT)]);

  // Air Density Correction Factor
  t = PRG_RDB(&AIRDENFACTOR[sensors8(MAT)]);

  // This is a hack! When idling for a long time, the intake air temperature usually rises
  // resulting in the intake air to be warm. When it gets too warm, the idling will be rough.
  // By tricking the msavr to think that the air ain't that hot, it won't lean the mixture
  // at idle rpm.
  // Suggested value of airden_ignore=62

  // We don't fiddle with the airden at cranking & warmup
  //  if (!(engine.status & (_BV(crank) | _BV(warmup))) && (t < config.airden_ignore)) {

  if (t < config.airden_ignore) {
    // normal running mode, but with low air density
    t = linear_interp(8, 14, config.airden_ignore, t, engine.rpm);
  }

  corr.air = t;

  // Interpolate throttle position
  // this makes it really smooth to install the tps
  // Should the tps be scaled 0..255 or 0..100% ?
  engine.tps = linear_interp_allow_swap(config.tps_low, config.tps_high, 0, 255, sensors8(TPS));

  // Do we want battery interpolation?
  // if the resistors in the voltagedivider are unprecise, or the vref is 
  // a little off, the batt_tmp in "calc_total_enrichment" will be wrong
  engine.batt = batt_cal(sensors8(VBATT), config.batt_cal);
}


void baro_correction(uint8_t t) {

	// Barometric correction enabled?
	if (config.config13 & _BV(BARO_CORRECTION)) {
		uint8_t val;

		if (config.config11 & _BV(MAP_SENSOR))
			val = PRG_RDB(&BAROFAC4250[t]);
		else
			val = PRG_RDB(&BAROFAC4115[t]);

		// If the mcu resets while the engine is running, we don't believe 
		// 20 kPa being the atmospheric pressure.
		if ( (val < (config.baro - config.dbaro)) || 
		     (val > (config.baro + config.dbaro)) ) {
#warning "MegaTune wants the raw sensor value and will not discover the cheat!"
			val = config.baro;
			cli();
			engine.status_ext |= _BV(baro_problem);
			sei();
		}
		corr.baro = val;
	} else {
		corr.baro = 100; // no, not enabled
	}
}

#ifdef ALS
/***************************************************************************/
/*** AntiLagSystem-induced enrichment                                    ***/
/***************************************************************************/
#warning ALS code is VERY EXPERIMENTAL and it is NOT TESTED yet!!! Use it at your own risk!
void als_enrich(void)
{
	uint8_t my_status, rpm, tps, als_sw_off;
	cli();
	my_status=engine.status;
	rpm=engine.rpm;
	tps=engine.tps;
	als_sw_off=als_deactivated;
	sei();
	if(my_status & _BV(crank) ||
		als_sw_off ||
		rpm <= config.als_lowrpm ||
		tps > config.als_maxtps)
	{
		cli();
		engine.status_ext &= ~_BV(als_active);
		sei();
		corr.als=0;
	}
	else
	{
		cli();
		engine.status_ext |= _BV(als_active);
		sei();
		corr.als=config.als_rich;
	}
}
#endif // ALS

/***************************************************************************/
/*** calc engine.rpm_vhr and engine.rpm                                  ***/
/***************************************************************************/
void calc_rpm(void) {
	uint16_t rpm, rpmk, rpm_p;
	uint32_t rpm_p32, rpmk32;

	// make sure the bytes are fetched atomically
	cli();
	rpm_p32 = engine.rpm_period;
	sei();

#ifdef BENCHMARK
	if (rpm_p32 < benchstats.lowest_rpm_p)
		benchstats.lowest_rpm_p = rpm_p32;
	if (rpm_p32 > benchstats.highest_rpm_p)
		benchstats.highest_rpm_p = rpm_p32;
#endif


	if((uint16_t)(rpm_p32 >> 16)){	// very small RPM
		cli();
		engine.rpm_vhr = 100;
		sei();
		engine.rpm = 1;
		return;
	}
    
	// rpm_p is the number of 4 usec's in a tach-event
	rpm_p = rpm_p32 & 0xFFFF;

	// rpm_p below config.crank_minper [16 usec resolution] is considered fake
	// this also prevents division by zero below
	if(rpm_p < ((uint16_t)(config.crank_minper) << 2)){	// abnormally small period
		if (rpm_p) {  // silly trigger 
			engine.rpm = 255;			// 25500 RPM
			cli();
			engine.rpm_vhr = 25500;	// fake value
			sei();
		} else {	// seems there were no valid rpm_p values yet
		// Note that the writeback of rpm_vhr is atomic. 
		// Since we do not call this from interrupt, only from userspace:
		// cli()-sei() is fine (no SREG saving needed)
			engine.rpm = 0;
			cli();
			engine.rpm_vhr = 0;
			sei();
		}
		return;
	}


	rpmk = (config.rpmk[0] << 8) | config.rpmk[1];  // I think I'll set it 3000 (0x0BB8) for my 4 cyl
	
	// MIK: The following multiplications must be type casted to 32bit, else the
	// upper two bytes of the result will be truncated.

	// 25 since rpm_p is in 4 usec not 100. 100 because we want real rpm, not [100 rpm] resolution
	rpmk32 = mult16_16(rpmk, 25*100);
	
	// rpm_hr was the number of rotations in 6 sec (real RPM/10)
	// no division by zero, because of checks above
	rpmk = rpmk32 / rpm_p;	// real rpm
	cli();
	engine.rpm_vhr = rpmk; 
	sei();
	// division by 100: * 41 / 4096
	// think about rounding. What do we want? Maybe rpmk += 50 before truncating?
	rpm = div4096(mult16_8(rpmk, 41));	// faster and more safe than trying 8*8 mult

	if ((rpm >> 8) != 0)
		engine.rpm = 255; // 25.500 rpm should be enough for anybody (Bill Gates '81)
	else
		engine.rpm = rpm & 0xFF;

}


/***************************************************************************/
/*** cranking                                                            ***/
/***************************************************************************/
void cranking(void) {
  uint8_t my_status;

  cli();
  my_status = engine.status;
  my_status |= _BV(crank);
  my_status &= ~( _BV(startw) | _BV(warmup) );
  engine.status = my_status;
  sei();
}


/**************************************************************************
* primepulse
* Call under cli()
***************************************************************************/
void primepulse(void) {
	uint16_t primep_cold, primep_hot, pw;
	uint8_t primep;

	primep = config.primep;
	if (primep == 0){ 
		// BEWARE: fill_all_squirters() not called (injpwmt OK?)
		// TODO: is that OK?
		return;
	}
	
    // don't prime if the engine is flooded
	if (engine.tps >= FLOOD_CLEAR)
		return;
  
	primep_cold = progressive_delay(primep);
	primep_hot = mult16_8(primep_cold, config.primep_temp_scaling) >> 8;
	
	pw = linear_interp16( COLD_TEMP, WARM_TEMP, primep_cold, primep_hot, engine.coolant );
      
	// do we want battery correction?

	pwm_dc_fill();

	if (pw == 0) { // if the pulsewidth turned out to be zero
		return;
	}
	
	fill_all_squirters(pw);

	/* the primepulse is handled similarly as any injection event-sequence 
	* we are in the initial cli() condition, so no need for special care
	*/

	readTMR();

#ifdef GENBOARDv3
	injector_start_all();
#else
	injector_start(&inj_port[0], startinj1, INJAGOPWM, INJASTOP);
	injector_start(&inj_port[1], startinj2, INJBGOPWM, INJBSTOP);
#endif
	alarm2core();
}	// primepulse()

/*@}*/

