/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Michael Kristensen mik at caffrey.dk]        *
***********************************************************
*  multitooth decoding                                    *
\*********************************************************/

#ifndef __MULTITOOTH_C__
#define __MULTITOOTH_C__

#include "global.h"
#include "eventqueue.h"
#include "fuelcontrol.h"
#include "benchmark.h"

#include <inttypes.h>

#define TW_TRIG    0
#define TW_MISSING 1

#define invalidate_rpm_period() engine.rpm_period = 0xFFFFffff
#define isvalid_rpm_period() ((uint16_t)engine.rpm_period) != 0xffff

#define TOOTHREL_ERROR (0)
#define TOOTHREL_MISSING (1)
#define TOOTHREL_SINGLE (2)

// true if b > 1.5 * ave (NOTE: avehalf must be ave/2)
#define significantly_bigger(b, ave, avehalf) ((b > ave) && ((b - ave) > avehalf))

/*
 * single tooth return TOOTHREL_SINGLE
 * missing tooth return TOOTHREL_MISSING
 * error return TOOTHREL_ERROR
 *
 * config examples:
 * 60-2: config.toothrel_normal=0x65 or config.toothrel_normal=0x76
 * N-1:  config.toothrel_normal=0x65 or config.toothrel_normal=0x76
 *
 * 60-2: config.toothrel_missing=0xB5
 * N-1:  config.toothrel_missing=0x00
 */
static inline uint8_t tooth_relative_length( uint16_t b, uint16_t ave, uint16_t avehalf )
{
	uint8_t t8;

	if( (config.primary_trigger & _BV(trigger_multitooth_hairy)) == 0 ){
		// traditional behaviour, simple detection:
		if( significantly_bigger( b, ave, avehalf )){
			return TOOTHREL_MISSING;
		}
		return TOOTHREL_SINGLE;
	}

	// upper and lower threshold for normal tooth length
	t8 = config.toothrel_normal ;
	if( b >= ave ){
		uint16_t diff = b - ave;
		// longer than average
		// difference is LE than - say - 6/16
		if( diff <= mult16_8( ave, (t8>>4)) >> 4 ){ 
			return TOOTHREL_SINGLE;
		}

 		/* TODO: optimize for 16bit comparisons.
 		 * in some branches
		 * return TOOTHREL_ERROR will be needed if ave>... or so
		 * if ( ave > 65535/27*8) return TOOTHREL_ERROR;
		 */
		// from now on, t8 is the upper and lower threshold for missing tooth length
		t8 = config.toothrel_missing ;
		if( diff < mult16_8( ave, 4 + (t8 & 0x0F)) >> 3 ){
			// but diff is LT - say - 9 / 8, so not long enough for a missing tooth
			return TOOTHREL_ERROR;
		}

		if( diff > mult16_8( ave, 12 + (t8 >> 4)) >> 3 ){
			// or diff  is GT - say - 23/8, so too long
			return TOOTHREL_ERROR;
		}
		return TOOTHREL_MISSING;

	}
	// likely single tooth 
	// difference is LE than - say - 5/16
	if( (ave - b) <= mult16_8( ave, (t8 & 0x0F)) >> 4 ){
		// note: b > mult16_8( ave, (t8 & 0x0F)) >> 4  would be nicer, but more config docs
		return TOOTHREL_SINGLE;
	}

	// OOPS .. too low
	return TOOTHREL_ERROR;
}


/*
	Tooth detection.
	Used from trigger interrupt.
  
	Bit definitions of trigger:
    bit 0: trigger
    bit 1: missing tooth
	
	tooth wheel in sync when etime.tooth_cnt < 128
	
	tooth_detection operates on 16 bit period instead of 32 bit
*/
static inline uint8_t tooth_detection(uint16_t period16, uint32_t capturedtime) {
	uint8_t trigger = 0, tooth_cnt, c_trigger_tooth, c_another_trigger_tooth;
	uint16_t etime_atp, etime_atp2;

	tooth_cnt = etime.tooth_cnt;
	
	etime_atp = etime.average_tooth_period;
	etime_atp2 = etime_atp >> 1;	// half
	if (tooth_cnt < 128) {	// wheel_in_sync
		uint8_t toothlen;
		uint8_t ctw = config.tooth_wheel;
		tooth_cnt++;
		toothlen = tooth_relative_length(period16, etime_atp, etime_atp2);
		if (toothlen == TOOTHREL_ERROR){
			// start sync sequence all from the beginning
			etime.tooth_cnt = 255;
			return 0;
		}

		if( toothlen == TOOTHREL_MISSING ) {
			if (tooth_cnt == ctw) {
				engine.rpm_period = capturedtime - etime.prev_missing_tooth;
				etime.prev_missing_tooth = capturedtime;
				engine.status_ext |= _BV(new_rpm);  // start a rpm calculation
				trigger = _BV(TW_MISSING); // trigger was 0
				tooth_cnt = 0;
#ifdef ALLOW_SLOPPY_TOOTHWHEEL
			// another chance:
			// Marcell's toothwheel sometimes counts 60-3 instead of 60-2
			// this means trigger came later _LAST_TIME_:
			// therefore it _was_ timing 6 degrees later, well..
			// why should we lose the next 2 events ?
			} else if( tooth_cnt == ctw - 1  ){
				engine.rpm_period = capturedtime - etime.prev_missing_tooth;
				etime.prev_missing_tooth = capturedtime;
				engine.status_ext |= _BV(new_rpm);  // start a rpm calculation
				trigger = _BV(TW_MISSING); // trigger was 0
				tooth_cnt = 0;
#ifdef BENCHMARK
				// increase wheel_err so it gets noticed
				benchstats.wheel_err++;
#endif	// BENCHMARK
#endif // ALLOW_SLOPPY_TOOTHWHEEL
			} else {
				// something went wrong, result of the tooth wheel decoding cannot be trusted
				// leave last known good engine.rpm_period as it is
//				invalidate_rpm_period();
				engine.status_trig |= _BV(wheel_error);
#ifdef BENCHMARK
				benchstats.wheel_err++;
#endif	// BENCHMARK
				// reenter sync-mode, the only sane thing to do
				etime.tooth_cnt = 255;
				return 0;
			}
		}
	} else { // NOT wheel_in_sync
#define TOOTH_WHEEL_NOISE 3
#define TOOTH_WHEEL_CNT   5
#define TOOTH_WHEEL_LISTEN (TOOTH_WHEEL_NOISE + TOOTH_WHEEL_CNT)

		engine_is_running(); //turn on fuelpump etc.
// do a rather long sync sequence to avoid getting wild rpm readings during the first rotation

		if (tooth_cnt < (255 - TOOTH_WHEEL_LISTEN)) {
			// initialization done, find the missing tooth
			if (tooth_relative_length(period16, etime_atp, etime_atp2) == TOOTHREL_MISSING ) {
				// uint32_t t32;
				engine.status_trig &= ~_BV(wheel_error);

				// TODO: use timing data from more teeth
				// the >> 1 is because period16 is for 2 teeth at this point
				// would be good for 36-1 only, since 60-2 is 3 teeth !
				// t32 = mult16_8( period16, config.another_trigger_tooth ) >> 1;
				// engine.rpm_period = ((t32 >> 16) ? 65535 : (uint16_t) t32);
				// we just make sure that TW_TRIG not fired in this case:

				invalidate_rpm_period();
	
				etime.prev_missing_tooth = capturedtime;
				trigger |= _BV(TW_MISSING);
				tooth_cnt = 0;
			} 
			// else stay unsynced tooth_cnt kept constant at 254 - TOOTH_WHEEL_LISTEN
			// note that we do update etime.average_tooth_period before return
		} else {  // assume the first pulses are noise
			--tooth_cnt;
			if (tooth_cnt > 255 - TOOTH_WHEEL_NOISE){
				invalidate_rpm_period();
				// we don't even avarage at this point.
				// (although a bogus 0xffff * 4usec  would decay fast)
				etime.average_tooth_period = period16;	// doesn't hurt to save it
				etime.tooth_cnt = tooth_cnt;
				return 0;
			} // else note the average_tooth_period averaging
		} 
	}

/* test cases naming:
Tat_t another_trigger_tooth=at trigger_tooth=t
note: RPM always applied at 0 anyway, see start of this function
T30_0 RPM: 30 	TRIG: 0,30
T30_1		TRIG: 1,31
T30_29		TRIG: 29,59
T20_0 RPM: 20,40 TRIG: 0,20,40
T20_1		TRIG: 1,21,41
T20_19		TRIG: 19,39,59
T15_0 RPM: 15,30,45 TRIG: 0,15,30,45
T15_1		TRIG: 1,16,31,46
T15_14		TRIG: 14,29,44,59
T9_0 RPM: 9,18,27 TRIG: 0,9,18,27
T9_1		TRIG: 1,10,19,28
T9_8		TRIG: 8,17,26,35
 ... is another_trigger_tooth=0 ever applied ?
T0_0 RPM: -  
T0_1
T0_11
*/
	etime.tooth_cnt = tooth_cnt;

	// the trigger-tooth can be adjusted in config
	c_trigger_tooth = config.trigger_tooth;
	c_another_trigger_tooth = config.another_trigger_tooth;
	if (c_another_trigger_tooth){
		while( tooth_cnt > c_another_trigger_tooth){
			tooth_cnt -= c_another_trigger_tooth;
		}
	// take care, tooth_cnt destroyed now, subtracted N * c_another_trigger_tooth
	// tooth_cnt range:  0 <= tooth_cnt <= c_another_trigger_tooth INCLUSIVE range !!!

		if( tooth_cnt == c_another_trigger_tooth ) {
			engine.rpm_period = capturedtime - etime.prev_missing_tooth;
			etime.prev_missing_tooth = capturedtime;
			engine.status_ext |= _BV(new_rpm);  // start a rpm calculation
			// NO trigger = _BV(TW_MISSING);
		}
	}

	// see "tooth_cnt range" note above
	if( ( tooth_cnt == c_trigger_tooth || (c_trigger_tooth == 0 && tooth_cnt == c_another_trigger_tooth))
		&& isvalid_rpm_period() ) {
		trigger |= _BV(TW_TRIG);
		// updating engine.rpm_period and "start a rpm calculation" happens elsewhere
	}

	// first take the half, than add them! Overflow is more concern than 4 usec precision
	etime.average_tooth_period = etime_atp2 + (period16 >> 1);
	return trigger;
}

#endif
