/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Michael Kristensen, mik@caffrey.dk]          *
***********************************************************
*  -- History --                                          *
*    release 0.1 - 06-03-2003                             *
*            http://caffrey.dk/megasquirt                 *
*    initial Motorola version by Bowling & Grippo (2002)  *
*            http://www.bgsoflex.com/megasquirt.html      *
\*********************************************************/


/*
  ##### VE AUTO TUNING #####
  
  How it works:
    The function 've_learn' is called every 0.1 second. When N 
    [rpm,kpa,ego correction]-samples have been gathered, the mean
    and variance of all N samples are calculated.

    The mean specifies which VE entry to modify and the variance gives a 
    figure of how much the data can be trusted. A small variance allows
    the VE entry to be changed much.

    The 'config.ve_learn_speed' is the speed that the VE table is changed. Even
    though the variance is 0, the VE table shouldn't change instantaniously.

*/

   /*!
   * \addtogroup g_fuel_calc Fuel Calculation Routines
   */
/*@{*/

#include "global.h"
#include "helpers.h"
#include "ve.h"
#include "adc.h"
#include "fuelcalc.h"
#include "ego.h"
#include "storage.h"
#include "benchmark.h"
#include "knock.h"
#include "iac.h"
#include "boostcontrol.h"

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <avr/pgmspace.h>
#include <string.h>

#include "arithm.c"
#include "edis.c"

volatile extern struct engine_t engine;

// these are global until we are out of the debug phase...
struct ve_at_t tune_me;
struct ve_stat_res_t res;
struct search_table_t kpa_idx, rpm_idx;

#ifndef NOIGN
/* calculates dwell in 4 usec precision - simple interpolation.
If EDIS is enabled in config, dwell should be the SAW commandlength,
*/
static __inline__ uint16_t calc_dwell(void)
{
	uint16_t dw;
	uint8_t s;
	if(config.ign_out & _BV(IGN_DUMMYDWELL)){
	// for anything above 13.8V we use the 0th dwellvalue (should we decrease for higher voltages?)
	// this is very convenient, the user normally need to set only 1 value
	dw = config.ign_dwell14 << 4;	// from 64 usec we go to 4 usec	
	s = engine.batt;
	// in fuelcalc.c injector pulsewidth correction batt_tmp 
	// 164 (appr.14V?) and 61 ADC readings seem to be typical high-low values
	if(s<BATT_14V) {	// lower Battery voltage, correction applies
		dw += ((BATT_14V-s) * config.ign_dwell6)>>3;
		// 118-61 = 54
		// ign.dwell6==x (slope) means x*(54*4)/8 usec dwellincrease at VBatt=appr 6V
		// set it to a value similar to dwell[0].
		// example:
		// config.ign_dwell14 = 16
		// means 16 * 64 usec dwell at 14Vbatt or above
		// config.ign_dwell6 = 17
		// means dwell extended _with_ _additional_ 17 * 27 usec at 6Vbatt
	}
	return dw;
	
	} 
	// EDIS
	return edis_sawlen();
}

static inline uint8_t triggertooth(uint8_t adv)
{
/*
  precise ignition trigger - the hack way. I want something working
  now, then the fine java models can contribute when they are finished :)

  Calculate trigger tooth N, and the time from tooth N to spark 'trig2spark'.
  In the multitooth handler, the d(tooth_period) is calculated:
    tooth_period{N+1} = 2*tooth_period{N} - tooth_period{N-1}.

  when tooth_period{N+1} is found, the event is scheduled to obtain a
  precise firing of the spark.

  Example:
  Say trigger tooth is defined (in config) to be 7 teeth BTDC (10
  degrees/tooth), and an advance of 36 degrees is requested.
  The trigger tooth will be N=3 and 4 degrees remains. 
  In the multitooth handler, IGNDEACT is scheduled to
  t =  k * 4/10 * tooth_period{N+1}, where k is computed in a similar
  way as done in trig2spark().

*/
	uint8_t toothw, trigtooth, degr;
	uint16_t tdcdelay;

	tdcdelay = ((uint16_t)config.ign_tdcdelay)<<1;	// going 1/4 crankdeg resolution

	if(adv > tdcdelay)
		return 0; // argh, the user configured higher advance than ign_tdcdelay?

	// note:  this is a redefinition of tooth_wheel_twidthX.
	// for this purpose it is easier if the unit is [deg] or [deg/4]
	// TODO: take care of the missing tooth
	trigtooth = 0;
	toothw = config.tooth_wheel_twidth1 << 2;  // going 1/4 degree resolution
	while ((tdcdelay - adv) > toothw) {
		tdcdelay -= toothw;
		trigtooth++;
	}

	degr = tdcdelay - adv;

    return 0;
}

// we get ignition advance and return value for engine.trig2spark scaler 
// which will be used in trigger.c
static inline uint16_t trig2spark(uint8_t adv)
{
	uint8_t ncyl2;
	uint16_t adv16, t16;
	
	adv16 = (uint16_t)adv;
	t16 = ((uint16_t)config.ign_tdcdelay)<<1;	// going 1/4 crankdeg resolution
	if(t16 > adv16)
		t16 -= adv16;	// this is ign_tdcdelay*2 - ignadv
	else // argh, the user configured higher advance than ign_tdcdelay?
		return 0;

	// trig2spark usage: time = engine.rpm_period * engine.trig2spark / 4096;
	// note: t16 is max 510 (/4 = 127.5 crankdegree). 
	// the rpm_period depends on ncyl (180 degree for 4 cyl, 90 deg for 8 cyl)
	// for 8 cyl t16=360 (90 degree) scales the period with *= 1.0 (return 4096)
	// for 4 cyl t16=360 (90 degree) scales the period with *= 0.5 (return 2048)
	// reference period value is 0xffff: RPM=57.22 for 8cyl

	// not ncyl, but 2 * ncyl
	ncyl2 = ((config.config11 & 0xF0) >> 3) + 2;
	// reusing variable:
	adv16 = mult16_8(t16, ncyl2);	// max 510 * 15 * 2
	return mult16_8(adv16, 182)>>8;
}

// calculate the time needed for the crank to rotate d/4 degrees
// todo: limits checking (d < 4096)
uint16_t degree2time(uint16_t d)
{
	uint8_t ncyl2;
	uint16_t t16;
	
	// usage: time = engine.rpm_period * degree2time(n) / 4096;
	// note: t16 is max 510 (/4 = 127.5 crankdegree). 
	// the rpm_period depends on ncyl (180 degree for 4 cyl, 90 deg for 8 cyl)
	// for 8 cyl t16=360 (90 degree) scales the period with *= 1.0 (return 4096)
	// for 4 cyl t16=360 (90 degree) scales the period with *= 0.5 (return 2048)
	// reference period value is 0xffff: RPM=57.22 for 8cyl

	// not ncyl, but 2 * ncyl
	ncyl2 = ((config.config11 & 0xF0) >> 3) + 2;
	// reusing variable:
	t16 = mult16_8(d, ncyl2);	// max 510 * 15 * 2
	return mult16_8(t16, 182)>>8;
}
#endif // NOIGN

/*
 * blends between MAP and TPS:
 * TPS at low RPM
 * MAP at high RPM
 * blends in between 
 
 http://www.vems-group.org/wiki/index.php?page=GenBoard%2FUnderDevelopment%2FAlphaN 
*/
static inline uint8_t engine_kpa_tps_blend8( void ){

	uint8_t rpm, rpm1, rpm2;
	
	rpm = engine_rpm8();
	rpm1 = config.hybrid_rpm_a;
	rpm2 = config.hybrid_rpm_m;
	/*
	if(rpm < rpm1 ){
		// with "<" never triggered if hybrid_rpm_a == 00
		// Note that the linear_interp() uses <= which means 
		// this could be accidentally triggered at 0..100 RPM
		// but cranking pulsewidth is used anyway
		return engine_tps8();
	}
	
	if( rpm >= rpm2 ){
		// never triggered if hybrid_rpm_m == ff
		return engine_kpa8();
	}	
	
	*/	
	return linear_interp(rpm1, rpm2, engine_tps8(), engine_kpa8(), rpm );
}


/**
 * Calculates indexes to be used for table-lookups.
 * Uses engine_kpa_tps_blend8()
 * (result is also saved to engine.kpa_tps_blend
 * for usage in table-lookup() )
 * to consider AlphaN mode when necessary.
 */
void search_rpm_kpa_index(void) {
	uint8_t t8 = engine_kpa_tps_blend8();
	engine.kpa_tps_blend = t8;
	search_table(ve.kpa_range, sizeof(ve.kpa_range), t8, &kpa_idx);
  search_table(ve.rpm_range, sizeof(ve.rpm_range), engine.rpm, &rpm_idx);
}

/***************************************************************************/
/*** table lookup                                                        ***/
/***************************************************************************/
void table_lookup(void) {
  uint8_t ve_11, ve_12, ve_21, ve_22;
  uint8_t t_11, t_12, t_21, t_22;
  uint16_t w_11, w_12, w_21, w_22;	// weights of the 2x2 box
  uint16_t rpm_weight, kpa_weight, rpm_weight255, kpa_weight255;
  uint16_t rx21_rl, rx21_kl, dist_rpm, rpm_vhr;
  uint8_t ku;
  uint16_t t1, t2;
  uint8_t rpm_u;

	rpm_u = ve.kpa_range[kpa_idx.uindex]; // rpm_u holds upper kpa for a while
	rx21_kl = div_65535_x( rpm_u - ve.kpa_range[kpa_idx.lindex] );

	// example values are for 5000, 5800, 6000 RPM
	rpm_u = ve.rpm_range[rpm_idx.uindex]; // rpm_u now holds upper rpm (=60)
	rx21_rl = div_65535_x( rpm_u - ve.rpm_range[rpm_idx.lindex] ); // =65535/10=6554

	// calculate rpm-distance and column weight:
	dist_rpm = rpm_u * 25;	// (=1500) upper rpm value in 4 RPM resolution (not distance yet)
	rpm_vhr = engine.rpm_vhr;// (=5800) no cli-sei, since rpm_vhr filled from userspace
	rpm_vhr >>= 2;	// (=1450) 4 RPM resolution
	if (rpm_vhr < dist_rpm){ // (=true)
		uint16_t rx21_rl_d25;

		dist_rpm -= rpm_vhr;	// (=50) now distance in 4 RPM units
		// rpm_weight255 = mult16_8(rx21_rl, ve.rpm_range[rpm_idx.uindex] - engine.rpm);
		rx21_rl_d25 = div25(rx21_rl); // (=262)
		// rx21_rl_16 = rx21_rl >> 4;
		// rx21_rl_d25 = mult16_8(rx21_rl_16, 163) >> 8; 	
		rpm_weight255 = mult16_16(rx21_rl_d25, dist_rpm); //(=262*50=13100=appr.65536/5)  cannot overflow
	} else
		rpm_weight255 = 0;

	rpm_weight = 0xffff - rpm_weight255;

#ifdef BOOSTCONTROL
	// TODO: add gear selection
#define BOOSTGEAR 0
	t_21 = boost.boostmaxtarget_table[ BOOSTGEAR * VE_SIZE_RPM + rpm_idx.lindex ];
	t_22 = boost.boostmaxtarget_table[ BOOSTGEAR * VE_SIZE_RPM + rpm_idx.uindex ];
	w_21 = t_21 * (rpm_weight >> 8);
	w_22 = t_22 * (rpm_weight255 >> 8);
	// take care of overflow (price is cycles. >>8 before addition would cost precision):
	engine.boostmax_target = ( (w_21 >> 4) + (w_22 >> 4) )>>4;
#endif // BOOSTCONTROL

	// calculate kpa-distance and row weight:
	ku = ve.kpa_range[kpa_idx.uindex];
	t_11 = engine.kpa_tps_blend;
	if (t_11 < ku)
		kpa_weight255 = mult16_8(rx21_kl, ku - t_11);
	else
		kpa_weight255 = 0;

	kpa_weight = 0xffff - kpa_weight255;
	
	// There is little point in 16 bit weights: 
	// the precision of the values is _more_ important
	
	// low rpm, low kpa
	w_11 = mult16_16(kpa_weight255, rpm_weight255)>>16;
	// high rpm, low kpa
	w_12 = mult16_16(kpa_weight255, rpm_weight)>>16;
	// low rpm, high kpa
	w_21 = mult16_16(kpa_weight, rpm_weight255)>>16;
	// high rpm, high kpa
	w_22 = mult16_16(kpa_weight, rpm_weight)>>16;


	// table indexes:
	t_11 = VE_SIZE_RPM * kpa_idx.lindex + rpm_idx.lindex;
	t_12 = VE_SIZE_RPM * kpa_idx.lindex + rpm_idx.uindex;
	t_21 = VE_SIZE_RPM * kpa_idx.uindex + rpm_idx.lindex;
	t_22 = VE_SIZE_RPM * kpa_idx.uindex + rpm_idx.uindex;

	// using 16 bit variables here (reusing names, don't get confused)
	t1 = ve.ve_table[ t_11 ];
	t2 = ve.ve_table[ t_12 ];
	dist_rpm = ve.ve_table[ t_21 ];
	rpm_vhr = ve.ve_table[ t_22 ];

	// the 16 bit weighted VE value (weights are 8 bit here):
	corr.ve_hp = scalar_4x4(t1, t2, dist_rpm, rpm_vhr,  w_11>>8, w_12>>8, w_21>>8, w_22>>8) >> 8;
  
  // the old fraction calc 
  // was carriless: not as "precise" as should be
  // check TestScalarMult() class in JTune CVS.
  // or imagine that all 16 bits are averaged independently.
  // the error is exactly 0/4, 1/4, 2/4 or 3/4 of the upper byte's LSB
  
  // lambdacorr table
  ve_11 = ve.lambda[ t_11 ];
  ve_12 = ve.lambda[ t_12 ];
  ve_21 = ve.lambda[ t_21 ];
  ve_22 = ve.lambda[ t_22 ];

  corr.lambda = scalar_4x4(w_11, w_12, w_21, w_22, ve_11, ve_12, ve_21, ve_22) >> 16;

#ifndef NOIGN

	// variables reused, but this is ign, not ve (do not get confused)
	ve_11 = ign.adv_table[ t_11 ];
	ve_12 = ign.adv_table[ t_12 ];
	ve_21 = ign.adv_table[ t_21 ];
	ve_22 = ign.adv_table[ t_22 ];
	// reusing uint8_t ku and uint16_t rpm_weight, kpa_weight variables, don't be confused
	ku = scalar_4x4(w_11, w_12, w_21, w_22, ve_11, ve_12, ve_21, ve_22) >> 16;

//TODO: we need an engine.looked_up_ignadv
//	engine.ignadv = ku;

#ifdef ALS
	if (engine.status_ext & _BV(als_active)) // we don't want to knock-knock when we can do BANG-BANG! ;-)
		if (ku <= config.als_ignretard)
			ku=0;
		else
			ku -= config.als_ignretard;
	else
#endif // ALS
#ifdef KNOCK_8101
	{
		t_11 = knock_cyl[engine.igncount].knock_adjust;

		if (t_11 >= ku)
			ku = 0;
		else 
			ku -= t_11;

	}

#endif // KNOCK_8101
	; // needed for close the 'else'
#ifdef DANGEROUS_KNOCK_TESTING
	t1 = ku + engine.ignadv_extra;
	if (t1 > 0xff)
		ku = 0xff;
	else
		ku = (uint8_t)t1;
#endif

	// apply adjustments from idle control
	ku = idle_advance_adjust(ku);

	// mdp menu command can force a pulsewidth in 64 usec resolution
	if (debug.forced_adv) {
		ku = debug.forced_adv;
	} else if (engine.rpm <= config.cranking_thres) {
		ku = config.ign_crank_advance;
	}

	t1 = trig2spark(ku);
	t2 = calc_dwell();

	// beware of Murphy...
	cli();
	engine.ignadv = ku;
	engine.trig2spark = t1;
	engine.dwell = t2;

#ifdef KNOCK_8101
// can it be guaranteed that the correct knock_adjust is applied
// in case the main loop is heavily loaded? ve_igncount will tell:
	knock_stats.ve_igncount = engine.igncount;  // debug
#endif
	sei();	

#endif	// NOIGN

}




/***************************************************************************/
/*** ve_entry_adjust                                                     ***/
/***************************************************************************/
/*
  ve_entry_adjust modifies a single VE entry pointed to by [rpm,kpa]. The 
  amount that the entry shall be changed is delivered in the adjust argument.

  ex: VE[i,j]=80 is increased by 3%, adjust will have the value 
      adjust =3%/100*256 + 128 = 135
      Weight is a scaling factor of how much of the change is applied.
*/
void ve_entry_adjust(uint8_t rpm_idx, uint8_t kpa_idx, uint8_t weight, uint8_t adjust) {
	// adjust biased at EGOCORR_BASE (==128)
	if ((adjust != EGOCORR_BASE) && (weight > config.ve_learn_min_weight)) {
		uint16_t ve_entry, ve_new, t, w, diff, ulimit, llimit;
		uint8_t ve_reference, deviation, adj;

		// fetch current VE value and fraction
		ve_entry = ve.ve_table[ VE_SIZE_RPM * kpa_idx + rpm_idx ];
    
		// calculate the candidate for the final VE value,
		// the lower 8 bits of ve_entry is the fraction

		// calculate the relative change of ve_entry and scale the result by weight and config.ve_learn_speed
		if (adjust >= EGOCORR_BASE)
			adj = adjust - EGOCORR_BASE;
		else 
			adj = EGOCORR_BASE - adjust;

		t = mult16_8(ve_entry, adj) >> 8;	// t <= ve_entry / 2
		w = weight * config.ve_learn_speed;
		
		// diff can never be more than appr. ve_entry / 2
		diff = mult16_16(t, w) >> 16;
	
		if (adjust >= EGOCORR_BASE)
			if(diff >= 0-ve_entry) // check for overflow
				ve_new = 0xffff;
			else
				ve_new = ve_entry + diff;
		else
			ve_new = ve_entry - diff;	// never overflows

		// limit the change, eg. compare ve_new with reference VE
		ve_reference = ve.ref_table[ VE_SIZE_RPM * kpa_idx + rpm_idx ];

		deviation = (ve_reference * config.ve_learn_limit) >> 8;

		// handle overflow
		t = ve_reference + deviation;
		ulimit = (t>0xff) ? 0xffff : t<<8 | 0xff;  // this is a lot of fuel!

		llimit = (ve_reference - deviation) << 8;
		// note that deviation <= ve_reference, so not needed:
		// llimit = (t>0xff) ? 0x0000 : t<<8; // "underflow"

		if (ve_new > ulimit) {   // ve_new too large
			ve_new = ulimit;
		} else if (ve_new < llimit) {   // ve_new too small
			ve_new = llimit;
		}

		// store VE value and fraction
		ve.ve_table[ VE_SIZE_RPM * kpa_idx + rpm_idx ] = ve_new;
	}
}


/***************************************************************************/
/*** ve_auto_tune                                                        ***/
/***************************************************************************/
void ve_auto_tune(struct ve_at_t *a) {
	struct search_table_t kpa, rpm;
	uint16_t drpm, dkpa;
	uint8_t dist_rpm, dist_kpa;
	uint8_t rpm_weight, kpa_weight, rpm_weight255, kpa_weight255;
	uint16_t t16;
	uint8_t adj, weight;
	uint8_t rpm_u;

	// calc the Constant/(x2-x1) hint values for the no-div operation
	rpm_u = ve.kpa_range[kpa_idx.uindex]; // rpm_u holds upper kpa for a while
	dkpa = div_65535_x( rpm_u - ve.kpa_range[kpa_idx.lindex] );

	// example values are for 5000, 5800, 6000 RPM
	rpm_u = ve.rpm_range[rpm_idx.uindex]; // rpm_u now holds upper rpm (=60)
	drpm = div_65535_x( rpm_u - ve.rpm_range[rpm_idx.lindex] ); // =65535/10=6554

	// first find the 4 VE entries that should be evaluated
	search_table(ve.rpm_range, sizeof(ve.rpm_range), a->rpm, &rpm);
	search_table(ve.kpa_range, sizeof(ve.kpa_range), a->kpa, &kpa);

	// distance from higher VE-point to actual [rpm, kpa]
	dist_rpm = ve.rpm_range[rpm.uindex] - a->rpm;
	dist_kpa = ve.kpa_range[kpa.uindex] - a->kpa;

	// calculate weight
	// weight is defined as 'weight related to higher VE-point' (just like ex. dist_rpm)
	// but we preserve the weight from the lower point as well in weight255 variables:
	rpm_weight255 = mult16_8(drpm, dist_rpm) >> 8;
	kpa_weight255 = mult16_8(dkpa, dist_kpa) >> 8;

	rpm_weight = 255 - rpm_weight255;
	kpa_weight = 255 - kpa_weight255;

	// adjust wrt confidence
	// adjust is 128 based, small confidence brings adj towards 128
	t16 = tune_me.adjust * tune_me.confidence;
	t16 -= (uint16_t)(128 * tune_me.confidence);
	adj = 128 + (t16 >> 8);

	if (config.ve_learn_conf & _BV(ve_learn_2x2)) {
		// modify all four elements of the {rpm,kpa}-box

		// low rpm, low kpa
		weight = (uint16_t)(rpm_weight255 * kpa_weight255) >> 8;
		a->w00 = weight;
		ve_entry_adjust(rpm.lindex, kpa.lindex, weight, adj);

		// low rpm, high kpa
		weight = (uint16_t)(rpm_weight255 * kpa_weight) >> 8;
		a->w01 = weight;
		ve_entry_adjust(rpm.lindex, kpa.uindex, weight, adj);

		// high rpm, low kpa
		weight = (uint16_t)(rpm_weight * kpa_weight255) >> 8;
		a->w10 = weight;
		ve_entry_adjust(rpm.uindex, kpa.lindex, weight, adj);

		// high rpm, high kpa
		weight = (uint16_t)(rpm_weight * kpa_weight) >> 8;
		a->w11 = weight;
		ve_entry_adjust(rpm.uindex, kpa.uindex, weight, adj);

	} else {
		// only modify nearest element
		a->w00 = a->w01 = a->w10 = a->w11 = 0;

		if (rpm_weight > 127) {	// rpm.uindex column
			if (kpa_weight > 127) {// kpa.uindex line
				weight = (uint16_t)(rpm_weight * kpa_weight) >> 8;
				a->w11 = weight;
				ve_entry_adjust(rpm.uindex, kpa.uindex, weight, adj);
			} else {				// kpa.lindex line
				weight = (uint16_t)(rpm_weight * kpa_weight255) >> 8;
				a->w10 = weight;
				ve_entry_adjust(rpm.uindex, kpa.lindex, weight, adj);
			}
		} else {					// rpm.lindex column
			if (kpa_weight > 127) {// kpa.uindex line
				weight = (uint16_t)(rpm_weight255 * kpa_weight) >> 8;
				a->w01 = weight;
				ve_entry_adjust(rpm.lindex, kpa.uindex, weight, adj);
			} else {				// kpa.lindex line
				weight = (uint16_t)(rpm_weight255 * kpa_weight255) >> 8;
				a->w00 = weight;
				ve_entry_adjust(rpm.lindex, kpa.lindex, weight, adj);
			}
		}
	}
}


/***************************************************************************/
/*** ve_check_limit                                                      ***/
/***************************************************************************/
// check if rpm & kpa is within limits for safe modification of VE entry
// returns 1 if ok, 0 if power is too high
uint8_t ve_check_limit(uint8_t rpm, uint8_t kpa) {
	uint8_t power;
	power = (rpm * kpa) >> 8;

	if (power < config.ve_learn_max_power)
		return 1;
	else
		return 0;

}


/***************************************************************************/
/*** ve_learn                                                            ***/
/***************************************************************************/
void ve_learn(void) {
	static uint8_t pfill;
	static uint8_t sample_count;
	static struct ve_learn_stat_t learn_stat;

	// store [rpm,kpa,ego] in circular buffer
	learn_stat.rpm[pfill] = engine.rpm;
	learn_stat.kpa[pfill] = engine_kpa8();
	learn_stat.ego[pfill] = corr.ego;

	if (++pfill >= STAT_SIZE)
		pfill=0;

	// we don't want acceleration/deacceleration/low_coolant_temp to interfere
//	if (!( (engine.status_ext & _BV(o2_closed_loop)) && (corr.warm == 0)) ) {
	if (!( (engine.status_ext & _BV(o2_closed_loop)) && 
	       (engine.coolant >= config.ve_learn_coolant)) ) {

		// just clear the good_sample_count, the O2 correction is reset anyways when
		// tpsaen or tpsden are activated
		sample_count = 0;

	} else {

		// ok, still in closed loop, increase number of good samples
		sample_count++;
    
		if (sample_count >= (VE_LEARN_SAMPLES+VE_LEARN_SKIP_SAMPLES)) {
			//      struct ve_stat_res_t res;
			//      sample_count = VE_LEARN_SAMPLES+VE_LEARN_SKIP_SAMPLES;

			ve_calc_stat((uint8_t*)&learn_stat.rpm, &res.rpm);
			ve_calc_stat((uint8_t*)&learn_stat.kpa, &res.kpa);
			ve_calc_stat((uint8_t*)&learn_stat.ego, &res.ego);
      
			if (ve_check_limit(res.rpm.mean, res.kpa.mean)) {
				uint8_t conf_rpm, conf_kpa, conf_ego;
				//      struct ve_at_t tune_me;

				tune_me.rpm = res.rpm.mean;
				tune_me.kpa = res.kpa.mean;
				tune_me.adjust = res.ego.mean;

				conf_rpm = ve_variance_scale(res.rpm.variance, config.ve_learn_rpm_scale);
				conf_kpa = ve_variance_scale(res.kpa.variance, config.ve_learn_kpa_scale);
				conf_ego = ve_variance_scale(res.ego.variance, config.ve_learn_ego_scale);

				tune_me.cr = conf_rpm;
				tune_me.ck = conf_kpa;
				tune_me.ce = conf_ego;

				// the confidence depends on variation of all three variables
				tune_me.confidence = mult16_8(conf_rpm * conf_kpa, conf_ego) >> 16;

				if (tune_me.confidence != 0) {
					if ((config.ve_learn_conf & _BV(ve_learn_sim)) == 0)
						ve_auto_tune(&tune_me);
				}
			}
			sample_count=VE_LEARN_SKIP_SAMPLES;
		}
	}
}


/***************************************************************************/
/*** ve_variance_scale                                                   ***/
/***************************************************************************/
uint8_t ve_variance_scale(uint8_t val, uint8_t scale) {
  uint16_t t = scale * val;
  
  if (t >= 255)
    return 0;
  else 
    return 255 - t;
}


/***************************************************************************/
/*** ve_calc_stat                                                        ***/
/***************************************************************************/
void ve_calc_stat(uint8_t *data, struct res_t *res) {
	uint8_t i;
	uint16_t mean, var, sum, d;
	
	// calculate mean
	mean = 0;
	for (i=0; i<STAT_SIZE; i++) {
		mean += data[i];
	}
	mean = mean / STAT_SIZE;
	res->mean = mean;    // store the mean

	// calculate variance
	sum = 0;
	for (i=0; i<STAT_SIZE; i++) {
		var = sum;
		d = data[i] - mean;
		sum = var + (mult16_16(d, d) & 0xffff);
		if (sum < var) {  // overflow
			var = 0xffff;
			break;
		}
	}
	var = var / STAT_SIZE;
	res->variance = var;	// store the variance
}


/***************************************************************************/
/*** sum_square                                                          ***/
/***************************************************************************/
/*void sum_square(int8_t val, uint16_t *sum) {
  uint16_t t = *sum + val * val;

  if (t < *sum)
    *sum = 0xFFFF;  // overflow
  else 
    *sum = t;
}
*/





/***************************************************************************/
/***************************************************************************/
/*** general purpose functions                                           ***/
/***************************************************************************/
/***************************************************************************/




/***************************************************************************/
/*** copyVEtoVEref                                                       ***/
/***************************************************************************/
void copyVEtoVEref(void) {
  uint8_t i;
  for (i=0; i<VE_SIZE_RPM*VE_SIZE_KPA; i++) {
    ve.ref_table[i] = (ve.ve_table[i] + 0x80) >> 8;
  }
}

/***************************************************************************/
/*** copyVEreftoVE                                                       ***/
/***************************************************************************/
void copyVEreftoVE(void) {
  uint8_t i;
  for (i=0; i<VE_SIZE_RPM*VE_SIZE_KPA; i++) {
    ve.ve_table[i] = (ve.ref_table[i] << 8) | 0x7F;
  }
}


/***************************************************************************/
/*** interpolateVE                                                       ***/
/***************************************************************************/
// when using a 16x16 VE table, this function will fill the blanks if the 
// table is configured as a 8x8
void interpolateVE(void) {
  

}

/*@}*/

